#include "CIOObject.h"

#include "utils/dbg/StackGuard.h"
#include "utils/dbg/CallStack.h"
#include "utils/dbg/Assert.h"
#include "ui/CLog.h"
#include "threads/ThreadLocalStorage.h"
#include "threads/Threads.h"
#include "threads/CIOTimer.h"

#include "threads/IOCP.h"
#include "Orts.h"

#include <vector>
#include <queue>

CIOCriticalSection CIOObject::g_globalLockTimer;  // L2CacheD 0x026846B0
volatile LONG CIOObject::s_objectInTimer;         // L2CacheD 0x02684698

typedef std::vector<CIOTimer> TimerVector;
typedef std::priority_queue<CIOTimer, TimerVector> TimerQueue;
static TimerQueue g_timerQueues[ThreadIndex_MaxCount];  // L2CacheD 0x26846F8

CIOObject::CIOObject()
    : MemoryObject()
{
}

// L2CacheD 0x00427E80
CIOObject::~CIOObject()
{
}

// L2CacheD 0x00426C50
void CIOObject::TimerExpired(int /*id*/)
{
}

// L2CacheD 0x00414B40
void CIOObject::OnTimer(int id)
{
    setInTimer(false);
    TimerExpired(id);
}

// L2CacheD 0x0047F4A0
void CIOObject::OnTimerCallback(int id)
{
    guard(L"void CIOObject::OnTimerCallback(int id)", L"void CIOObject::OnTimerCallback(int id) (exit)", TEXT(__FILE__), __LINE__);

    if (::PostQueuedCompletionStatus(IOCP::g_hCompletionPort, id, (ULONG_PTR)this, NULL) == FALSE)
    {
        EASSERT();
        Release(__FILE__, __LINE__, ORT_TIMER, 0);
    }

    unguard();
}

// L2CacheD 0x00401070
void CIOObject::OnWaitCallback()
{
}

// L2CacheD 0x0047F540
void CIOObject::OnIOCallback(BOOL /*success*/, DWORD transferred, LPOVERLAPPED /*lpOverlapped*/)
{
    guard(L"void CIOObject::OnIOCallback(BOOL success, DWORD transferred, LPOVERLAPPED lpOverlapped)", L"void CIOObject::OnIOCallback(BOOL success, DWORD transferred, LPOVERLAPPED lpOverlapped) (exit)", TEXT(__FILE__), __LINE__);

    OnTimer(transferred);
    Release(__FILE__, __LINE__, ORT_TIMER, 0);

    unguard();
}

// L2CacheD 0x00414B50
BOOL CIOObject::PostObject(DWORD id)
{
    return ::PostQueuedCompletionStatus(IOCP::g_hCompletionPort, id, (ULONG_PTR)this, NULL);
}

// L2CacheD 0x0047F5D0
bool CIOObject::SetShortTimeStamp(long* pTimeVar, DWORD nAdditional)
{
    guard(L"bool CIOObject::SetShortTimeStamp(long* pTimeVar, DWORD nAdditional)", L"bool CIOObject::SetShortTimeStamp(long* pTimeVar, DWORD nAdditional) (exit)", TEXT(__FILE__), __LINE__);

    long now = nAdditional + ::GetTickCount();
    if (now <= *pTimeVar)
    {
        unguard();
        return false;
    }

    ::InterlockedExchange(pTimeVar, now);

    unguard();

    return true;
}

// L2CacheD 0x00481710
void CIOObject::TimerDispatch()
{
    guard(L"void CIOObject::TimerDispatch()", L"void CIOObject::TimerDispatch() (exit)", TEXT(__FILE__), __LINE__);

    DWORD tick = ::GetTickCount();
    uint32_t threadIndex = tls::GetCurrentThreadIndex();

    TimerQueue& timerQueue = g_timerQueues[threadIndex];

    while (!timerQueue.empty())
    {
        const CIOTimer& top = timerQueue.top();
        if (top.m_dwTime > tick)
        {
            break;
        }

        int id = top.m_id;
        CIOObject* object = top.m_pObject;
        timerQueue.pop();

        ::InterlockedDecrement(&CIOObject::s_objectInTimer);

        if (threadIndex == ThreadIndex_ListenThread)
        {
            object->OnTimerCallback(id);
        }
        else
        {
            object->OnIOCallback(true, id, NULL);
        }
    }

    unguard();
}

#define MAXIMUM_WAIT_OBJECTS 64

// L2CacheD 0x00482650
bool CIOObject::RegisterWait(HANDLE handle)
{
    guard(L"bool CIOObject::RegisterWait(HANDLE handle)", L"bool CIOObject::RegisterWait(HANDLE handle) (exit)", TEXT(__FILE__), __LINE__);

    g_globalLockTimer.Enter(0, 0);
    if (Threads::g_vHandle.size() >= MAXIMUM_WAIT_OBJECTS)
    {
        g_globalLockTimer.Leave(0, 0);
        unguard();
        return false;
    }

    Threads::g_vHandle.push_back(handle);
    Threads::g_vObject.push_back(this);

    g_globalLockTimer.Leave(0, 0);

    unguard();

    return true;
}

// L2CacheD 0x004828C0
bool CIOObject::AddTimer(unsigned int time, int id)
{
    guard(L"bool CIOObject::AddTimer(unsigned time, int id)", L"bool CIOObject::AddTimer(unsigned time, int id)", TEXT(__FILE__), __LINE__);

    if (time < 80)
    {
        // small timers limit
        static int s_count = 0;
        if (s_count++ < 30)
        {
            g_winlog.Add(LOG_ERROR, L"dwTime[%d] < 80 !!!!!!!", time);
            CallStack::ReportCallStack();
        }
        time = 80;
    }

    ::InterlockedIncrement(&CIOObject::s_objectInTimer);

    AddRef(__FILE__, __LINE__, ORT_TIMER);

    setInTimer(true);

    uint32_t threadIndex = tls::GetCurrentThreadIndex();
    DWORD timeout = time + ::GetTickCount();

    TimerQueue& timerQueue = g_timerQueues[threadIndex];
    timerQueue.push(CIOTimer(this, timeout, id));

    unguard();

    return true;
}
