#pragma once

#include "Orts.h"
#include "threads/CRWLock.h"
#include "threads/CSPointer.h"

#include <map>

#include <utils/cstdint_support.h>

template <typename T>
class CSPointer;

template <typename T>
class CObjectMap
{
public:
    // if returns true, ForEach exits from loop and returns curr object
    // In original L2 code ForEach is probably a macro
    typedef bool (*Predicate)(T* object);

public:
    CObjectMap();   // L2CacheD T=CItem, 0x0040CC40
    ~CObjectMap();  // L2CacheD T=CItem, 0x0040CCE0

    CSPointer<T> FindObject(uint32_t objectId) const;                                // L2CacheD T=CItem, 0x00425640
    void Insert(uint32_t id, T* object, const char* caller = NULL, int line = -1);   // L2CacheD T=CItem, 0x00425A30
    void DeleteObject(uint32_t objectId, const char* caller = NULL, int line = -1);  // L2CacheD T=CItem, 0x00425B10; T=CUser, 0x004035B0
    T* ForEach(Predicate predicate);                                                 // L2CacheD always inlined, probably a macro
    void DeleteAll(Predicate predicate);                                             // L2CacheD always inlined, probably a macro

private:
    typedef std::map<int, T*> ObjMap;

    ObjMap m_objects;
    mutable CRWLock m_lock;
    uint32_t m_last;
};

#include "utils/dbg/StackGuard.h"

// L2CacheD T=CItem, 0x0040CC40
template <typename T>
CObjectMap<T>::CObjectMap()
    : m_objects()
    , m_lock()
    , m_last(1)
{
    guard(L"CObjectMap()", L"CObjectMap() (exit)", TEXT(__FILE__), __LINE__);
    unguard();
}

// L2CacheD T=CItem, 0x0040CCE0
template <typename T>
CObjectMap<T>::~CObjectMap()
{
    guard(L"~CObjectMap()", L"~CObjectMap() (exit)", TEXT(__FILE__), __LINE__);

    m_lock.WriteLock();

    for (ObjMap::iterator it = m_objects.begin(); it != m_objects.end(); ++it)
    {
        ObjMap::value_type& pair = *it;
        T* object = pair.second;
        object->Release(__FILE__, __LINE__, ORT_CACHED, false);
    }

    m_objects.clear();
    m_lock.WriteUnlock();

    unguard();
}

// L2CacheD T=CItem, 0x00425640
template <typename T>
CSPointer<T> CObjectMap<T>::FindObject(uint32_t objectId) const
{
    guard(L"CSPointer<T> FindObject(uint32_t id)", L"CSPointer<T> FindObject(int id) (exit)", TEXT(__FILE__), __LINE__);

    m_lock.ReadLock();

    CSPointer<T> result;
    ObjMap::const_iterator it = m_objects.find(objectId);
    if (it != m_objects.end())
    {
        result.SetObject(it->second, ORT_CACHED);
    }
    m_lock.ReadUnlock();

    unguard();

    return result;
}

// L2CacheD T=CItem, 0x00425A30
template <typename T>
void CObjectMap<T>::Insert(uint32_t objectId, T* object, const char* caller, int line)
{
    guard(L"void Insert(uint32_t objectId, T* p, char *caller = NULL, int line = -1)", L"void Insert(int id, T* p, char *caller = NULL, int line = -1) (exit)", TEXT(__FILE__), __LINE__);

    m_lock.WriteLock();

    ObjMap::iterator it = m_objects.find(objectId);
    if (it == m_objects.end())
    {
        object->AddRef((caller == NULL) ? __FILE__ : caller, (line == -1) ? __LINE__ : line, ORT_CACHED);
        m_objects[objectId] = object;
    }

    m_lock.WriteUnlock();

    unguard();
}

// L2CacheD T=CItem, 0x00425B10
template <typename T>
void CObjectMap<T>::DeleteObject(uint32_t objectId, const char* caller, signed int line)
{
    guard(L"void DeleteObject(uint32_t id, char *caller=NULL, int line=-1)", L"void DeleteObject(int id, char *caller=NULL, int line=-1) (exit)", TEXT(__FILE__), __LINE__);

    m_lock.WriteLock();

    ObjMap::iterator it = m_objects.find(objectId);
    if (it != m_objects.end())
    {
        ObjMap::value_type& pair = *it;
        T* object = pair.second;
        object->Release((caller == NULL) ? __FILE__ : caller, (line == -1) ? __LINE__ : line, ORT_CACHED, false);

        // m_objects.erase(objectId);  FIXED
        m_objects.erase(it);
    }

    m_lock.WriteUnlock();

    unguard();
}

// L2CacheD always inlined, probably a macro
template <typename T>
T* CObjectMap<T>::ForEach(Predicate predicate)
{
    guard(L"FOR ALL OBJECT IN OBJECT MAP", L"FOR ALL OBJECT IN OBJECT MAP (exit)", TEXT(__FILE__), __LINE__);

    m_lock.ReadLock();

    for (ObjMap::iterator it = m_objects.begin(); it != m_objects.end(); ++it)
    {
        T* object = it->second;

        guard(L"in_code", L"in_code (exit)", TEXT(__FILE__), __LINE__);
        bool stop = predicate(object);
        unguard();

        if (stop)
        {
            return object;
        }
    }

    m_lock.ReadUnlock();

    unguard();
}
