#pragma once

#include "Orts.h"
#include "threads/CRWLock.h"
#include "threads/CSPointer.h"

#include <map>
#include <string>

template <typename T>
class CSPointer;

template <typename T>
class CStringObjectMap
{
public:
    // if returns true, ForEach exits from loop and returns curr object
    // In original L2 code ForEach is probably a macro
    typedef bool (*Predicate)(T* object);

public:
    CStringObjectMap();
    ~CStringObjectMap();

    CSPointer<T> FindObject(const std::wstring& name) const;
    void Insert(const std::wstring& name, T* object, const char* caller = NULL, int line = -1);
    void DeleteObject(const std::wstring& name, const char* caller = NULL, int line = -1);
    T* ForEach(Predicate predicate);
    void DeleteAll(Predicate predicate);

private:
    typedef std::map<std::wstring, T*> ObjMap;

    ObjMap m_objects;
    mutable CRWLock m_lock;
};

#include "utils/dbg/StackGuard.h"

template <typename T>
CStringObjectMap<T>::CStringObjectMap()
    : m_objects()
    , m_lock()
{
    guard(L"CStringObjectMap()", L"CStringObjectMap() (exit)", TEXT(__FILE__), __LINE__);
    unguard();
}

template <typename T>
CStringObjectMap<T>::~CStringObjectMap()
{
    guard(L"~CStringObjectMap()", L"~CStringObjectMap() (exit)", TEXT(__FILE__), __LINE__);

    m_lock.WriteLock();

    for (ObjMap::iterator it = m_objects.begin(); it != m_objects.end(); ++it)
    {
        ObjMap::value_type& pair = *it;
        T* object = pair.second;
        object->Release(__FILE__, __LINE__, ORT_CACHED, false);
    }

    m_objects.clear();
    m_lock.WriteUnlock();

    unguard();
}

template <typename T>
CSPointer<T> CStringObjectMap<T>::FindObject(const std::wstring& name) const
{
    guard(L"CSPointer<T> FindObject(const std::wstring& name)", L"CSPointer<T> FindObject(const std::wstring& name) (exit)", TEXT(__FILE__), __LINE__);

    m_lock.ReadLock();

    CSPointer<T> result;
    ObjMap::const_iterator it = m_objects.find(name);
    if (it != m_objects.end())
    {
        result.SetObject(it->second, ORT_CACHED);
    }
    m_lock.ReadUnlock();

    unguard();

    return result;
}

template <typename T>
void CStringObjectMap<T>::Insert(const std::wstring& name, T* object, const char* caller, int line)
{
    guard(L"void Insert(const std::wstring& name, T* p, char *caller = NULL, int line = -1)", L"void Insert(const std::wstring& name, T* p, char *caller = NULL, int line = -1) (exit)", TEXT(__FILE__), __LINE__);

    m_lock.WriteLock();

    ObjMap::iterator it = m_objects.find(name);
    if (it == m_objects.end())
    {
        object->AddRef((caller == NULL) ? __FILE__ : caller, (line == -1) ? __LINE__ : line, ORT_CACHED);
        m_objects[name] = object;
    }

    m_lock.WriteUnlock();

    unguard();
}

template <typename T>
void CStringObjectMap<T>::DeleteObject(const std::wstring& name, const char* caller, signed int line)
{
    guard(L"void DeleteObject(const std::wstring& name, char *caller=NULL, int line=-1)", L"void DeleteObject(const std::wstring& name, char *caller=NULL, int line=-1) (exit)", TEXT(__FILE__), __LINE__);

    m_lock.WriteLock();

    ObjMap::iterator it = m_objects.find(name);
    if (it != m_objects.end())
    {
        ObjMap::value_type& pair = *it;
        T* object = pair.second;
        object->Release((caller == NULL) ? __FILE__ : caller, (line == -1) ? __LINE__ : line, ORT_CACHED, false);

        // m_objects.erase(name);  FIXED
        m_objects.erase(it);
    }

    m_lock.WriteUnlock();

    unguard();
}

// L2CacheD always inlined, probably a macro
template <typename T>
T* CStringObjectMap<T>::ForEach(Predicate predicate)
{
    guard(L"FOR ALL OBJECT IN OBJECT MAP", L"FOR ALL OBJECT IN OBJECT MAP (exit)", TEXT(__FILE__), __LINE__);

    m_lock.ReadLock();

    for (ObjMap::iterator it = m_objects.begin(); it != m_objects.end(); ++it)
    {
        T* object = it->second;

        guard(L"in_code", L"in_code (exit)", TEXT(__FILE__), __LINE__);
        bool stop = predicate(object);
        unguard();

        if (stop)
        {
            return object;
        }
    }

    m_lock.ReadUnlock();

    unguard();
}
