#include "CIOObject.h"

#include "utils/dbg/StackGuard.h"
#include "utils/dbg/CallStack.h"
#include "utils/dbg/Assert.h"
#include "ui/CLog.h"
#include "threads/ThreadLocalStorage.h"
#include "threads/Threads.h"
#include "threads/CIOTimer.h"

#include "threads/IOCP.h"
#include "Orts.h"

#include <vector>
#include <queue>

CIOCriticalSection CIOObject::g_globalLockTimer;  // 0x01BD3C18
volatile LONG CIOObject::s_objectInTimer;         // 0x01BD3C0C

typedef std::vector<CIOTimer> TimerVector;
typedef std::priority_queue<CIOTimer, TimerVector> TimerQueue;
static TimerQueue g_timerQueues[ThreadIndex_MaxCount];  // 0x01BD3C38

CIOObject::CIOObject()
    : MemoryObject()
{
}

// 0x0040A450
CIOObject::~CIOObject()
{
}

// 0x00409250
void CIOObject::TimerExpired(int /*id*/)
{
}

// 0x00409260
void CIOObject::OnTimer(int id)
{
    setInTimer(false);
    TimerExpired(id);
}

// 0x004087B0
void CIOObject::OnTimerCallback(int id)
{
    guard(L"void CIOObject::OnTimerCallback(int id, bool bPost)");

    if (::PostQueuedCompletionStatus(IOCP::g_hCompletionPort, id, (ULONG_PTR)this, NULL) == FALSE)
    {
        EASSERT();
        Release(__FILE__, __LINE__, ORT_TIMER, 0);
    }

    unguard();
}

// 0x00409270
void CIOObject::OnWaitCallback()
{
}

// 0x00408840
void CIOObject::OnIOCallback(BOOL /*success*/, DWORD transferred, LPOVERLAPPED /*lpOverlapped*/)
{
    guard(L"void CIOObject::OnIOCallback(BOOL bSucess, DWORD dwTransferred, LPOVERLAPPED lpOverlapped)");

    OnTimer(transferred);
    Release(__FILE__, __LINE__, ORT_TIMER, 0);

    unguard();
}

// 0x00409280
BOOL CIOObject::PostObject(DWORD id)
{
    return ::PostQueuedCompletionStatus(IOCP::g_hCompletionPort, id, (ULONG_PTR)this, NULL);
}

// 0x004088C0
bool CIOObject::SetShortTimeStamp(long* pTimeVar, DWORD nAdditional)
{
    guard(L"CIOObject::SetShortTimeStamp()");

    long now = nAdditional + ::GetTickCount();
    if (now <= *pTimeVar)
    {
        unguard();
        return false;
    }

    ::InterlockedExchange(pTimeVar, now);

    unguard();

    return true;
}

// 0x0040A9F0
void CIOObject::TimerDispatch()
{
    guard(L"void CIOObject::TimerDispatch(bool bRootLoop)");

    DWORD tick = ::GetTickCount();
    uint32_t threadIndex = tls::GetCurrentThreadIndex();

    TimerQueue& timerQueue = g_timerQueues[threadIndex];

    while (!timerQueue.empty())
    {
        const CIOTimer& top = timerQueue.top();
        if (top.m_dwTime > tick)
        {
            break;
        }

        int id = top.m_id;
        CIOObject* object = top.m_pObject;
        timerQueue.pop();

        ::InterlockedDecrement(&CIOObject::s_objectInTimer);

        if (threadIndex == ThreadIndex_ListenThread)
        {
            object->OnTimerCallback(id);
        }
        else
        {
            object->OnIOCallback(true, id, NULL);
        }
    }

    unguard();
}

#define MAXIMUM_WAIT_OBJECTS 64

// 0x0040B810
bool CIOObject::RegisterWait(HANDLE handle)
{
    guard(L"BOOL CIOObject::RegisterWait(HANDLE handle)");

    g_globalLockTimer.Enter(0, 0);
    if (Threads::g_vHandle.size() >= MAXIMUM_WAIT_OBJECTS)
    {
        g_globalLockTimer.Leave(0, 0);
        unguard();
        return false;
    }

    Threads::g_vHandle.push_back(handle);
    Threads::g_vObject.push_back(this);

    g_globalLockTimer.Leave(0, 0);

    unguard();

    return true;
}

// 0x0040BA70
bool CIOObject::AddTimer(unsigned int time, int id)
{
    guard(L"bool CIOObject::AddTimer(unsigned dwTime, int id)");

    if (time < 80)
    {
        // small timers limit
        static int s_count = 0;
        if (s_count++ < 30)
        {
            g_winlog.Add(LOG_ERROR, L"dwTime[%d] < 80 !!!!!!!", time);
            CallStack::ReportCallStack();
        }
        time = 80;
    }

    ::InterlockedIncrement(&CIOObject::s_objectInTimer);

    AddRef(__FILE__, __LINE__, ORT_TIMER);

    setInTimer(true);

    uint32_t threadIndex = tls::GetCurrentThreadIndex();
    DWORD timeout = time + ::GetTickCount();

    TimerQueue& timerQueue = g_timerQueues[threadIndex];
    timerQueue.push(CIOTimer(this, timeout, id));

    unguard();

    return true;
}
