#include <WinSock2.h>

#include "network/CLogSocket.h"
#include "network/PacketUtils.h"
#include "network/CPacket.h"
#include "ui/CLog.h"
#include "network/CIOBuffer.h"
#include "utils/dbg/StackGuard.h"
#include "utils/dbg/CallStack.h"
#include "memory/Allocator.h"

// 0x00420850
long CLogSocket::s_nAlloc = ::InterlockedIncrement(&MemoryObject::s_refCountSlots);  // 0x01BD4FC0

// 0x0042D89C
const CLogSocket::PacketHandler CLogSocket::s_packetHandlers[CLogSocket::HANDLERS_NUMBER] = {&CLogSocket::packet0_SendMsg, &CLogSocket::packet1_CheckStatus, &CLogSocket::packet2_TestPacket};

// 0x0040D4E0
CLogSocket::CLogSocket(SOCKET socket, const PacketHandler* packetHandlers)
    : CIOSocket(socket)
    , m_socket(socket)
    , field_2(0)
    , m_packetHandlers(packetHandlers)
    , m_status(SocketStatus_Init)
    , m_packetSize(0)
{
    guard(L"CLogSocket::CLogSocket(SOCKET s, LogPacketFunc *aPacketTable) : CIOSocket(s)");

    unguard();
}

// 0x0040D560
CLogSocket::~CLogSocket()
{
}

// 0x0040D140
void* CLogSocket::operator new(size_t size)
{
    void* memory = g_allocator.allocate(size, 1, L"CLogSocket");

    ::InterlockedIncrement(&Allocator::g_allocatedObjects[CLogSocket::s_nAlloc]);
    Allocator::g_allocatedObjName[CLogSocket::s_nAlloc] = "CLogSocket";
    Allocator::g_allocatedMemory[CLogSocket::s_nAlloc] = sizeof(CLogSocket);

    return memory;
}

// 0x0040D190
void CLogSocket::operator delete(void* /*ptr*/)
{
    g_winlog.Add(LOG_ERROR, L"Never Call Oprator Delete!!! CLogSocket at file[%s], line[%d]", TEXT(__FILE__), __LINE__);
    CallStack::ReportCallStack();
}

// 0x0040D700
void CLogSocket::Release(const char* file, int line, ORTs /*nWorkType*/, bool /*flag*/)
{
    long refCountAfterRelease = ::InterlockedDecrement(&m_nRefCount);
    if (refCountAfterRelease > 0)
    {
        return;
    }

    if (refCountAfterRelease != 0)
    {
        if (refCountAfterRelease > -1)
        {
            return;
        }

        g_winlog.Add(LOG_ERROR, "Invalid Memory (%d) on CLogSocket(%d)", refCountAfterRelease, m_identity);
        if (file)
        {
            g_winlog.Add(LOG_ERROR, "           Caller : file[%s], line[%d]", file, line);
        }

        ReportOrt(0, 0);
        CallStack::ReportCallStack();

        return;
    }

    if (inTimer())
    {
        if (file)
        {
            g_winlog.Add(LOG_ERROR, "deleting in-timer-object![CLogSocket] at file[%s], line[%d]", file, line);
            ReportOrt(0, 0);
            ::InterlockedDecrement(&Allocator::g_allocatedObjects[CLogSocket::s_nAlloc]);
        }
        else
        {
            g_winlog.Add(LOG_ERROR, "deleting in-timer-object![CLogSocket]");
            ReportOrt(0, 0);
            ::InterlockedDecrement(&Allocator::g_allocatedObjects[CLogSocket::s_nAlloc]);
        }

        return;
    }

    this->~CLogSocket();

    g_allocator.deallocate(this);

    ::InterlockedDecrement(&Allocator::g_allocatedObjects[CLogSocket::s_nAlloc]);
}

// 0x0040D5F0
void CLogSocket::Send(const char* format, ...)
{
    guard(L"void CLogSocket::Send(const char* format, ...)");

    va_list va;
    va_start(va, format);

    if (m_status == SocketStatus_Closed)
    {
        unguard();
        return;
    }

    // CIOBuffer* buffer = new CIOBuffer();  // FIXED
    CIOBuffer* buffer = CIOBuffer::Alloc();
    int packetSize = PacketUtils::VAssemble(&buffer->m_Buffer[2], BUFFER_SIZE - 2, format, va);
    if (packetSize != 0)
    {
        packetSize += 2;
        buffer->m_Buffer[0] = char(packetSize & 0xFF);
        buffer->m_Buffer[1] = char((packetSize >> 8) & 0xFF);
        buffer->m_dwSize = packetSize;  // FIXED,
        Write(buffer);                  // moved to correct if-branch
    }
    else
    {
        g_winlog.Add(LOG_ERROR, L"%d: assemble too large packet. format %s", m_hSocket, format);
        buffer->Release();  // FIXED
    }

    unguard();
}

// 0x0040D570
void CLogSocket::OnClose()
{
    guard(L"void CLogSocket::OnClose()");

    m_status = SocketStatus_Closed;  // FIXED, added
    g_winlog.Sendable(false);
    g_winlog.Add(LOG_NORMAL, L"CLogSocket Close %x(%x)", m_socket, this);

    unguard();
}

// 0x0040D870
void CLogSocket::OnRead()
{
    guard(L"void CLogSocket::OnRead()");

    if (m_status == SocketStatus_Closed)
    {
        Close();
        unguard();
        return;
    }

    const int buffSize = m_pReadBuf->m_dwSize;

    if (buffSize <= 0)
    {
        Read(0);
        unguard();
        return;
    }

    int dwRead = 0;

    while (true)
    {
        // read size only
        while (true)
        {
            if (m_status != SocketStatus_Init)
            {
                break;
            }

            if (dwRead + 2 > buffSize)
            {
                Read(buffSize - dwRead);
                unguard();
                return;
            }

            m_packetSize = (m_pReadBuf->m_Buffer[dwRead + 1] << 8) + m_pReadBuf->m_Buffer[dwRead] - 2;
            if (m_packetSize <= 0 || m_packetSize > BUFFER_SIZE)
            {
                g_winlog.Add(LOG_ERROR, L"dbsocket %d: bad packet size %d", m_hSocket, m_packetSize);
                m_status = SocketStatus_Closed;
                Close();
                unguard();
                return;
            }

            dwRead += 2;
            m_status = SocketStatus_BytesRead;

            if (dwRead >= buffSize)
            {
                Read(0);
                unguard();
                return;
            }
        }

        if (m_status != SocketStatus_BytesRead)
        {
            Close();
            unguard();
            return;
        }

        if ((dwRead + m_packetSize) > buffSize)
        {
            Read(buffSize - dwRead);
            unguard();
            return;
        }

        // 2 bytes size, 3rd byte is packet type
        int packetId = m_pReadBuf->m_Buffer[dwRead];
        if (packetId >= HANDLERS_NUMBER)
        {
            g_winlog.Add(LOG_ERROR, L"unknown protocol %d", packetId);
            m_status = SocketStatus_Closed;
            Close();
            unguard();
            return;
        }

        CPacket* packet = CPacket::Alloc();

        packet->m_pSocket = this;
        packet->m_pBuf = m_pReadBuf;

        packet->m_pFunc = m_packetHandlers[packetId];

        AddRef(__FILE__, __LINE__, ORT_IO_RETAIN);
        m_pReadBuf->AddRef();

        ::InterlockedIncrement(&CPacket::g_nPendingPacket);

        packet->PostObject(dwRead);
        dwRead += m_packetSize;
        m_status = SocketStatus_Init;

        //        if (dwRead >= m_pReadBuf->m_dwSize)
        //        {
        //            Read(0);
        //            unguard();
        //            return;
        //        }
    }

    unguard();
}

// 0x0040D430
bool CLogSocket::packet0_SendMsg(CIOSocket* /*socket*/, const uint8_t* /*packet*/)
{
    guard(L"static bool LogSendMsgPacket(CLogSocket *socket, const unsigned char *packet)");
    unguard();

    return false;
}

// 0x0040D450
bool CLogSocket::packet1_CheckStatus(CIOSocket* /*socket*/, const uint8_t* packet)
{
    guard(L"static bool LogCheckStatus(CLogSocket *socket, const unsigned char *packet)");

    int status = 0;
    // PacketUtils::Disassemble(packet, L"d", &status); // FIXED
    PacketUtils::Disassemble(packet, "d", &status);

    unguard();

    return false;
}

// 0x0040D4C0
bool CLogSocket::packet2_TestPacket(CIOSocket* /*socket*/, const uint8_t* /*packet*/)
{
    guard(L"static bool LogTestPacket(CLogSocket *socket, const unsigned char *packet)");
    unguard();

    return false;
}
