#include "utils/dbg/CoreDump.h"

#include "utils/dbg/CallStack.h"
#include "utils/dbg/LockStack.h"
#include "utils/dbg/CExceptionInit.h"
#include "utils/Utils.h"
#include "threads/ThreadLocalStorage.h"
#include "threads/CThreadManager.h"
#include "network/CIOBufferPool.h"

#include <wchar.h>

CIOCriticalSection CoreDump::s_lock;  // 0x00441B94

// 0x00402F30
void CoreDump::CreateReport(EXCEPTION_POINTERS* exceptionInfo)
{
    HANDLE fileHandle = ::CreateFileW(CExceptionInit::s_logPath, GENERIC_WRITE, 0, 0, CREATE_ALWAYS, FILE_FLAG_WRITE_THROUGH | FILE_ATTRIBUTE_NORMAL, 0);
    if (fileHandle == INVALID_HANDLE_VALUE)
    {
        ::OutputDebugStringW(L"Error creating exception report");
        return;
    }

    ::SetEndOfFile(fileHandle);

    int ioPoolsSize = CIOBufferPool::getFreeSlotsNumberFromAllPools();

    CoreDump::WriteToFile(fileHandle, L"IOBufferPool - %d\r\n\r\n", ioPoolsSize);

    SYSTEMTIME systemTime;
    ::GetLocalTime(&systemTime);

    int second = systemTime.wSecond;
    int minute = systemTime.wMinute;
    int hour = systemTime.wHour;
    int day = systemTime.wDay;
    int month = systemTime.wMonth;
    int year = systemTime.wYear;
    DWORD threadId = ::GetCurrentThreadId();
    CoreDump::WriteToFile(fileHandle, L"[(%d) %04d/%02d/%02d %02d:%02d:%02d]: =======================\r\n", threadId, year, month, day, hour, minute, second);

    CONTEXT* context = exceptionInfo->ContextRecord;

    const wchar_t* moduleName = L"Unknown";
    wchar_t moduleFilePath[MAX_PATH];
    MEMORY_BASIC_INFORMATION memoryInfo;
    if (::VirtualQuery((void*)context->Eip, &memoryInfo, sizeof(memoryInfo)) && ::GetModuleFileNameW((HMODULE)memoryInfo.AllocationBase, moduleFilePath, sizeof(moduleFilePath)))
    {
        moduleName = Utils::GetFileName(moduleFilePath);
    }

    EXCEPTION_RECORD* exceptionRecord = exceptionInfo->ExceptionRecord;
    const char* exceptionStr = CoreDump::ExceptionToString(exceptionRecord->ExceptionCode);
    CoreDump::WriteToFile(fileHandle, L"%S in module %s at %04x:%08x.\r\n", exceptionStr, moduleName, context->SegCs, context->Eip);

    std::tm* startTime = std::localtime(&CExceptionInit::s_startTime);
    CoreDump::WriteToFile(fileHandle, L"start at %d/%d/%d %02d:%02d:%02d\r\n", startTime->tm_year + 1900, startTime->tm_mon + 1, startTime->tm_mday, startTime->tm_hour, startTime->tm_min, startTime->tm_sec);
    if ((exceptionRecord->ExceptionCode == EXCEPTION_ACCESS_VIOLATION) && (exceptionRecord->NumberParameters >= 2))
    {
        // ExceptionInformation[0] indicates if it was a read (0), write (1), or data execution
        // attempt (8).
        const bool readFrom = exceptionRecord->ExceptionInformation[0] == 0;

        wchar_t violationBuffer[1000];
        wsprintfW(violationBuffer, L"%s location %08x caused an access violation.\r\n", readFrom ? L"Read from" : L"Write to", exceptionRecord->ExceptionInformation[1]);
        CoreDump::WriteToFile(fileHandle, L"%s", violationBuffer);
    }

    CoreDump::WriteToFile(fileHandle, L"\r\n");
    CoreDump::WriteToFile(fileHandle, L"Registers:\r\n");
    CoreDump::WriteToFile(fileHandle, L"EAX=%08x CS=%04x EIP=%08x EFLGS=%08x\r\n", context->Eax, context->SegCs, context->Eip, context->EFlags);
    CoreDump::WriteToFile(fileHandle, L"EBX=%08x SS=%04x ESP=%08x EBP=%08x\r\n", context->Ebx, context->SegSs, context->Esp, context->Ebp);
    CoreDump::WriteToFile(fileHandle, L"ECX=%08x DS=%04x ESI=%08x FS=%04x\r\n", context->Ecx, context->SegDs, context->Esi, context->SegFs);
    CoreDump::WriteToFile(fileHandle, L"EDX=%08x ES=%04x EDI=%08x GS=%04x\r\n", context->Edx, context->SegEs, context->Edi, context->SegGs);
    CoreDump::WriteToFile(fileHandle, L"Bytes at CS:EIP:\r\n");

    for (int i = 0; i < 16; ++i)
    {
        const uint8_t* eip = (uint8_t*)context->Eip;
        CoreDump::WriteToFile(fileHandle, L"%02x ", eip[i]);
    }

    CoreDump::WriteToFile(fileHandle, L"\r\nStack dump:\r\n");

    int* stackPointer = (int*)context->Esp;
    int* stackBase = (int*)__readfsdword(4);
    if (stackBase > (stackPointer + 80))
    {
        stackBase = (stackPointer + 80);
    }

    int counter = 0;
    static wchar_t STACK_BUFFER[1000];  // 0x00455438
    wchar_t* iterator = STACK_BUFFER;
    const wchar_t* limit = &STACK_BUFFER[950];

    while (stackPointer + 1 <= stackBase)
    {
        if (!(counter % 8))
        {
            iterator += wsprintfW(iterator, L"%08x: ", stackPointer);
        }

        const wchar_t* spaceOrEoL = L" ";
        if (((++counter % 8) == 0) || (stackPointer + 2 > stackBase))
        {
            spaceOrEoL = L"\r\n";
        }

        iterator += wsprintfW(iterator, L"%08x%S", *stackPointer, spaceOrEoL);
        ++stackPointer;
        if (iterator > limit)
        {
            CoreDump::WriteToFile(fileHandle, L"%s", STACK_BUFFER);
            STACK_BUFFER[0] = 0;
            iterator = STACK_BUFFER;
        }
    }

    CoreDump::WriteToFile(fileHandle, L"%s", STACK_BUFFER);

    CoreDump::s_lock.Enter(L"Exception.cpp", __LINE__);
    for (uint32_t threadIndex = 0; threadIndex < CThreadManager::GetIOThreadNumber(); ++threadIndex)
    {
        if (threadIndex != tls::GetCurrentThreadIndex())
        {
            HANDLE threadId = CThreadManager::GetThread((ThreadIndices)threadIndex);
            ::SuspendThread(threadId);
        }
    }
    CoreDump::s_lock.Leave(L"Exception.cpp", __LINE__);

    if (!CThreadManager::isListenThread())
    {
        HANDLE listenThreadId = CThreadManager::GetThread(ThreadIndex_ListenThread);
        ::SuspendThread(listenThreadId);
    }

    if (!CThreadManager::isMainThread())
    {
        HANDLE mainThreadId = CThreadManager::GetThread(ThreadIndex_Main);
        ::SuspendThread(mainThreadId);
    }

    CoreDump::WriteToFile(fileHandle, L"\r\n\r\nGuardInfo : \r\n\r\n");
    for (uint32_t threadIndex = 0; threadIndex < CThreadManager::GetIOThreadNumber(); ++threadIndex)
    {
        CoreDump::WriteCallStack((ThreadIndices)threadIndex, fileHandle, L"IOThread");
    }

    CoreDump::WriteCallStack(ThreadIndex_SqlQueue, fileHandle, L"SqlQueueThread");  // FIXED: added for perf thread
    CoreDump::WriteCallStack(ThreadIndex_CliLog, fileHandle, L"CliLogThread");      // FIXED: added for perf thread
    CoreDump::WriteCallStack(ThreadIndex_PerfMonitor, fileHandle, L"PerfMonitor");  // FIXED: added for perf thread
    CoreDump::WriteCallStack(ThreadIndex_ListenThread, fileHandle, L"ListenThread");
    CoreDump::WriteCallStack(ThreadIndex_Main, fileHandle, L"MainThread");
    CoreDump::WriteToFile(fileHandle, L"GuardInfo end \r\n\r\n");
    CoreDump::WriteToFile(fileHandle, L"\r\n\r\n");

    ::CloseHandle(fileHandle);
}

// 0x00402E10
void CoreDump::WriteCallStack(ThreadIndices threadIndex, HANDLE fileHandle, const wchar_t* threadName)
{
    const wchar_t* label = (threadIndex == tls::GetCurrentThreadIndex()) ? L"ahehe" : L"good";

    CoreDump::WriteToFile(fileHandle, L"%s [%d] (%s):  ", threadName, threadIndex, label);

    int stackSize = CallStack::s_threadStackSizes[threadIndex];
    FuncList& funcNames = CallStack::s_functions[threadIndex];

    for (int i = 0; i < stackSize; ++i)
    {
        const wchar_t* funcName = funcNames[i];
        CoreDump::WriteToFile(fileHandle, L"%s -> ", funcName);
    }

    CoreDump::WriteToFile(fileHandle, L"    Lock Stack  : ");

    const LockStack& lockStack = g_stacks[threadIndex];
    for (int i = 0; i < lockStack.lockSize(); ++i)
    {
        const LockFrame& lockFrame = lockStack.frame(i);
        CoreDump::WriteToFile(fileHandle, L"%s(%d[%x])", lockFrame.m_lockName, lockFrame.m_line, lockFrame.m_deep);  // FIXED
    }

    CoreDump::WriteToFile(fileHandle, L"\r\n\r\n");
}

// 0x004026E0
void CoreDump::WriteToFile(HANDLE fileHandle, wchar_t* format, ...)
{
    static wchar_t localBuffer[20000];  // 0x00044B7F8

    va_list va;
    va_start(va, format);
    // vswprintf(localBuffer, format, va); FIXED
    vswprintf(localBuffer, 20000, format, va);

    char* str = Utils::UnicodeToMultibyte_WithoutAllocation(localBuffer);

    DWORD bytesWritten;
    ::WriteFile(fileHandle, str, strlen(str), &bytesWritten, 0);
}

// 0x00402740
const char* CoreDump::ExceptionToString(int exceptionCode)
{
    struct ExceptionName
    {
        int code;
        const char* name;
    };

    ExceptionName exceptions[24];
    exceptions[0].code = DBG_CONTROL_C;
    exceptions[0].name = "a Control-C";
    exceptions[1].code = DBG_CONTROL_BREAK;
    exceptions[1].name = "a Control-Break";
    exceptions[2].code = EXCEPTION_DATATYPE_MISALIGNMENT;
    exceptions[2].name = "a Datatype Misalignment";
    exceptions[3].code = EXCEPTION_BREAKPOINT;
    exceptions[3].name = "a Breakpoint";
    exceptions[4].code = EXCEPTION_ACCESS_VIOLATION;
    exceptions[4].name = "an Access Violation";
    exceptions[5].code = EXCEPTION_IN_PAGE_ERROR;
    exceptions[5].name = "an In Page Error";
    exceptions[6].code = STATUS_NO_MEMORY;
    exceptions[6].name = "a No Memory";
    exceptions[7].code = EXCEPTION_ILLEGAL_INSTRUCTION;
    exceptions[7].name = "an Illegal Instruction";
    exceptions[8].code = EXCEPTION_NONCONTINUABLE_EXCEPTION;
    exceptions[8].name = "a Noncontinuable Exception";
    exceptions[9].code = EXCEPTION_INVALID_DISPOSITION;
    exceptions[9].name = "an Invalid Disposition";
    exceptions[10].code = EXCEPTION_ARRAY_BOUNDS_EXCEEDED;
    exceptions[10].name = "a Array Bounds Exceeded";
    exceptions[11].code = EXCEPTION_FLT_DENORMAL_OPERAND;
    exceptions[11].name = "a Float Denormal Operand";
    exceptions[12].code = EXCEPTION_FLT_DIVIDE_BY_ZERO;
    exceptions[12].name = "a Float Divide by Zero";
    exceptions[13].code = EXCEPTION_FLT_INEXACT_RESULT;
    exceptions[13].name = "a Float Inexact Result";
    exceptions[14].code = EXCEPTION_FLT_INVALID_OPERATION;
    exceptions[14].name = "a Float Invalid Operation";
    exceptions[15].code = EXCEPTION_FLT_OVERFLOW;
    exceptions[15].name = "a Float Overflow";
    exceptions[16].code = EXCEPTION_FLT_STACK_CHECK;
    exceptions[16].name = "a Float Stack Check";
    exceptions[17].code = EXCEPTION_FLT_UNDERFLOW;
    exceptions[17].name = "a Float Underflow";
    exceptions[18].code = EXCEPTION_INT_DIVIDE_BY_ZERO;
    exceptions[18].name = "an Integer Divide by Zero";
    exceptions[19].code = EXCEPTION_INT_OVERFLOW;
    exceptions[19].name = "an Integer Overflow";
    exceptions[20].code = EXCEPTION_PRIV_INSTRUCTION;
    exceptions[20].name = "a Privileged Instruction";
    exceptions[21].code = EXCEPTION_STACK_OVERFLOW;
    exceptions[21].name = "a Stack Overflow";
    exceptions[22].code = 0xC0000142;  // STATUS_DLL_INIT_FAILED;
    exceptions[22].name = "a DLL Initialization Failed";
    exceptions[23].code = 0xE06D7363;
    exceptions[23].name = "a Microsoft C++ Exception";

    for (int i = 0; i < 24; ++i)
    {
        if (exceptionCode == exceptions[i].code)
        {
            return exceptions[i].name;
        }
    }

    return "Unknown exception type";
}
